import trio

from . import abc


class SlackAPI(abc.SlackAPI):
    """
    `asks curio` implementation of :class:`slack.io.abc.SlackAPI`
    """
    def __init__(self, *, session, **kwargs):
        self._session = session
        super().__init__(**kwargs)

    async def _request(self, method, url, headers, body):
        response = await self._session.request(method, url, headers=headers, data=body)
        return response.status_code, response.content, response.headers

    async def rtm(self, url=None, bot_id=None):
        raise NotImplementedError

    async def _rtm(self, url):
        raise NotImplementedError

    async def sleep(self, seconds):
        await trio.sleep(seconds)



import time
import json
import logging
import websocket

from . import abc
from .. import sansio, exceptions, events, methods

LOG = logging.getLogger(__name__)


class SlackAPI(abc.SlackAPI):
    """
    `requests` implementation of :class:`slack.io.abc.SlackAPI`
    """

    def __init__(self, *, session, **kwargs):
        self._session = session
        super().__init__(**kwargs)

    def _request(self, method, url, headers, body):

        response = self._session.request(method, url, headers=headers, data=body)
        return response.status_code, response.content, response.headers

    def _rtm(self, url):

        ws = websocket.create_connection(url)
        while True:
            event = ws.recv()
            if event:
                yield event
            else:
                self.sleep(0.5)

    def sleep(self, seconds):
        time.sleep(seconds)

    def _make_query(self, url, data=None, headers=None):

        while self._rate_limited and self._rate_limited > int(time.time()):
            self.sleep(1)

        try:
            url, body, headers = sansio.prepare_request(url=url, data=data, headers=headers,
                                                        global_headers=self._headers, token=self._token)
            status, body, headers = self._request('POST', url, headers, body)
            response_data = sansio.decode_response(status, headers, body)
        except exceptions.RateLimited as rate_limited:
            if self._retry_when_rate_limit:
                raise
            else:
                LOG.warning('Rate limited ! Waiting for %s seconds', rate_limited.retry_after)
                self._rate_limited = int(time.time()) + rate_limited.retry_after
                return self._make_query(url, data, headers)
        else:
            self._rate_limited = False
            return response_data

    def query(self, url, data=None, headers=None):
        """
        Query the slack API

        Args:
            url: :class:`slack.methods` or url string
            data: JSON encodable MutableMapping
            headers: Custom headers

        Returns:
            dictionary of slack API response data

        """
        return self._make_query(url, data, headers)

    def iter(self, url, data=None, headers=None, *, limit=200, iterkey=None, itermode=None):
        """
        Iterate over a slack API method supporting pagination

        Args:
            url: :class:`slack.methods` or url string
            data: JSON encodable MutableMapping
            headers:
            limit: Maximum number of results to return per call.
            iterkey: Key in response data to iterate over (required for url string).
            itermode: Iteration mode (required for url string) (one of `cursor`, `page` or `timeline`)

        Returns:
            Async iterator over `response_data[key]`

        """
        itervalue = None
        while True:
            data, iterkey, itermode = sansio.prepare_iter_request(url, data, iterkey=iterkey, itermode=itermode,
                                                                  limit=limit, itervalue=itervalue)
            response_data = self._make_query(url, data, headers)
            itervalue = sansio.decode_iter_request(response_data)
            for item in response_data[iterkey]:
                yield item

            if not itervalue:
                break

    def rtm(self, url=None, bot_id=None):
        """
        Iterate over event from the RTM API

        Args:
            url: Websocket connection url
            bot_id: Connecting bot ID

        Returns:
            :class:`slack.events.Event` or :class:`slack.events.Message`

        """
        while True:
            if not bot_id:
                auth = self.query(methods.AUTH_TEST)
                user_info = self.query(methods.USERS_INFO, {'user': auth['user_id']})
                bot_id = user_info['user']['profile']['bot_id']
                LOG.info('BOT_ID is %s', bot_id)
            if not url:
                url = (self.query(methods.RTM_CONNECT))['url']

            for data in self._rtm(url):
                event = events.Event.from_rtm(json.loads(data))
                if sansio.need_reconnect(event):
                    break
                elif sansio.discard_event(event, bot_id):
                    if event['type'] == 'reconnect_url':
                        url = event['url']
                else:
                    yield event



import json
import logging

from collections import defaultdict, MutableMapping
from . import exceptions

LOG = logging.getLogger(__name__)


class Event(MutableMapping):
    """
    MutableMapping representing a slack event coming from the RTM API or the Event API.

    Attributes:
        metadata: Metadata dispatched with the event when using the Event API
                  (see `slack event API documentation <https://api.slack.com/events-api#receiving_events>`_)
    """

    def __init__(self, raw_event, metadata=None):
        self.event = raw_event
        self.metadata = metadata

    def __getitem__(self, item):
        return self.event[item]

    def __setitem__(self, key, value):
        self.event[key] = value

    def __delitem__(self, key):
        del self.event[key]

    def __iter__(self):
        return iter(self.event)

    def __len__(self):
        return len(self.event)

    def __repr__(self):
        return 'Slack Event: ' + str(self.event)

    def clone(self):
        """
        Clone the event

        Returns:
            :class:`slack.events.Event`

        """
        return self.__class__(self.event, self.metadata)

    @classmethod
    def from_rtm(cls, raw_event):
        """
        Create an event with data coming from the RTM API.

        If the event type is a message a :class:`slack.events.Message` is returned.

        Args:
            raw_event: JSON decoded data from the RTM API

        Returns:
            :class:`slack.events.Event` or :class:`slack.events.Message`
        """
        if raw_event['type'].startswith('message'):
            return Message(raw_event)
        else:
            return Event(raw_event)

    @classmethod
    def from_http(cls, raw_body, verification_token=None, team_id=None):
        """
        Create an event with data coming from the HTTP Event API.

        If the event type is a message a :class:`slack.events.Message` is returned.

        Args:
            raw_body: Decoded body of the Event API request
            verification_token: Slack verification token used to verify the request came from slack
            team_id: Verify the event is for the correct team

        Returns:
            :class:`slack.events.Event` or :class:`slack.events.Message`

        Raises:
            :class:`slack.exceptions.FailedVerification`: when `verification_token` or `team_id` does not match the
                                                          incoming event's.
        """
        if verification_token and raw_body['token'] != verification_token:
            raise exceptions.FailedVerification(raw_body['token'], raw_body['team_id'])

        if team_id and raw_body['team_id'] != team_id:
            raise exceptions.FailedVerification(raw_body['token'], raw_body['team_id'])

        if raw_body['event']['type'].startswith('message'):
            return Message(raw_body['event'], metadata=raw_body)
        else:
            return Event(raw_body['event'], metadata=raw_body)


class Message(Event):
    """
    Type of :class:`slack.events.Event` corresponding to a message event type
    """

    def __repr__(self):
        return 'Slack Message: ' + str(self.event)

    def response(self, in_thread=None):
        """
        Create a response message.

        Depending on the incoming message the response can be in a thread. By default the response follow where the
        incoming message was posted.

        Args:
            in_thread (boolean): Overwrite the `threading` behaviour

        Returns:
             a new :class:`slack.event.Message`
        """
        data = {'channel': self['channel']}

        if in_thread:
            if 'message' in self:
                data['thread_ts'] = self['message'].get('thead_ts') or self['message']['ts']
            else:
                data['thread_ts'] = self.get('thread_ts') or self['ts']
        elif in_thread is None:
            if 'message' in self and 'thread_ts' in self['message']:
                data['thread_ts'] = self['message']['thread_ts']
            elif 'thread_ts' in self:
                data['thread_ts'] = self['thread_ts']

        return Message(data)

    def serialize(self):
        """
        Serialize the message for sending to slack API

        Returns:
            serialized message
        """
        data = {**self}
        if 'attachments' in self:
            data['attachments'] = json.dumps(self['attachments'])
        return data


class Router:
    """
    When receiving an event form the RTM API or the slack API it is useful to have a routing mechanisms for
    dispatching event to individual function/coroutine. This class provide such mechanisms for any
    :class:`slack.events.Event`.
    """
    def __init__(self):
        self._routes = defaultdict(dict)

    def register(self, event_type, handler, **detail):
        """
        Register a new handler for a specific :class:`slack.events.Event` `type` (See `slack event types documentation
        <https://api.slack.com/events>`_ for a list of event types).

        The arbitrary keyword argument is used as a key/value pair to compare against what is in the incoming
        :class:`slack.events.Event`

        Args:
            event_type: Event type the handler is interested in
            handler: Callback
            **detail: Additional key for routing
        """
        LOG.info('Registering %s, %s to %s', event_type, detail, handler.__name__)
        if len(detail) > 1:
            raise TypeError('Only one detail can be provided for additional routing')
        elif not detail:
            detail_key, detail_value = '*', '*'
        else:
            detail_key, detail_value = detail.popitem()

        if detail_key not in self._routes[event_type]:
            self._routes[event_type][detail_key] = {}

        if detail_value not in self._routes[event_type][detail_key]:
            self._routes[event_type][detail_key][detail_value] = []

        self._routes[event_type][detail_key][detail_value].append(handler)

    def dispatch(self, event):
        """
        Yields handlers matching the routing of the incoming :class:`slack.events.Event`.

        Args:
            event: :class:`slack.events.Event`

        Yields:
            handler
        """
        LOG.debug('Dispatching event %s', event.get('type'))
        for detail_key, detail_values in self._routes.get(event.get('type')):
            event_value = event.get(detail_key, '*')
            callbacks = detail_values.get(event_value, [])
            for callback in callbacks:
                yield callback



from enum import Enum

ROOT_URL = 'https://slack.com/api/'


class Methods(Enum):
    """
    Enumeration of available slack methods.

    Provides `iterkey` and `itermod` for :func:`SlackAPI.iter() <slack.io.abc.SlackAPI.iter>`.
    """
    # api
    API_TEST = (ROOT_URL + 'api.test', None, None)

    # apps.permissions
    APPS_PERMISSIONS_INFO = (ROOT_URL + 'apps.permissions.info', None, None)
    APPS_PERMISSIONS_REQUEST = (ROOT_URL + 'apps.permissions.request', None, None)

    # auth
    AUTH_REVOKE = (ROOT_URL + 'auth.revoke', None, None)
    AUTH_TEST = (ROOT_URL + 'auth.test', None, None)

    # bots
    BOTS_INFO = (ROOT_URL + 'bots.info', None, None)

    # channels
    CHANNELS_ARCHIVE = (ROOT_URL + 'channels.archive', None, None)
    CHANNELS_CREATE = (ROOT_URL + 'channels.create', None, None)
    CHANNELS_HISTORY = (ROOT_URL + 'channels.history', 'timeline', 'messages')
    CHANNELS_INFO = (ROOT_URL + 'channels.info', None, None)
    CHANNELS_INVITE = (ROOT_URL + 'channels.invite', None, None)
    CHANNELS_JOIN = (ROOT_URL + 'channels.join', None, None)
    CHANNELS_KICK = (ROOT_URL + 'channels.kick', None, None)
    CHANNELS_LEAVE = (ROOT_URL + 'channels.leave', None, None)
    CHANNELS_LIST = (ROOT_URL + 'channels.list', 'cursor', 'channels')
    CHANNELS_MARK = (ROOT_URL + 'channels.mark', None, None)
    CHANNELS_RENAME = (ROOT_URL + 'channels.rename', None, None)
    CHANNELS_REPLIES = (ROOT_URL + 'channels.replies', None, None)
    CHANNELS_SET_PURPOSE = (ROOT_URL + 'channels.setPurpose', None, None)
    CHANNELS_SET_TOPIC = (ROOT_URL + 'channels.setTopic', None, None)
    CHANNELS_UNARCHIVE = (ROOT_URL + 'channels.unarchive', None, None)

    # chat
    CHAT_DELETE = (ROOT_URL + 'chat.delete', None, None)
    CHAT_ME_MESSAGE = (ROOT_URL + 'chat.meMessage', None, None)
    CHAT_POST_EPHEMERAL = (ROOT_URL + 'chat.postEphemeral', None, None)
    CHAT_POST_MESSAGE = (ROOT_URL + 'chat.postMessage', None, None)
    CHAT_UNFURL = (ROOT_URL + 'chat.unfurl', None, None)
    CHAT_UPDATE = (ROOT_URL + 'chat.update', None, None)

    # dnd
    DND_END_DND = (ROOT_URL + 'dnd.endDnd', None, None)
    DND_END_SNOOZE = (ROOT_URL + 'dnd.endSnooze', None, None)
    DND_INFO = (ROOT_URL + 'dnd.info', None, None)
    DND_SET_SNOOZE = (ROOT_URL + 'dnd.setSnooze', None, None)
    DND_TEAM_INFO = (ROOT_URL + 'dnd.teamInfo', None, None)

    # emoji
    EMOJI_LIST = (ROOT_URL + 'emoji.list', None, None)

    # files.comments
    FILES_COMMENTS_ADD = (ROOT_URL + 'files.comments.add', None, None)
    FILES_COMMENTS_DELETE = (ROOT_URL + 'files.comments.delete', None, None)
    FILES_COMMENTS_EDIT = (ROOT_URL + 'files.comments.edit', None, None)

    # files
    FILES_DELETE = (ROOT_URL + 'files.delete', None, None)
    FILES_INFO = (ROOT_URL + 'files.info', None, None)
    FILES_LIST = (ROOT_URL + 'files.list', 'page', 'files')
    FILES_REVOKE_PUBLIC_URL = (ROOT_URL + 'files.revokePublicURL', None, None)
    FILES_SHARED_PUBLIC_URL = (ROOT_URL + 'files.sharedPublicURL', None, None)
    FILES_UPLOAD = (ROOT_URL + 'files.upload', None, None)

    # groups
    GROUPS_ARCHIVE = (ROOT_URL + 'groups.archive', None, None)
    GROUPS_CLOSE = (ROOT_URL + 'groups.close', None, None)
    GROUPS_CREATE = (ROOT_URL + 'groups.create', None, None)
    GROUPS_CREATE_CHILD = (ROOT_URL + 'groups.createChild', None, None)
    GROUPS_HISTORY = (ROOT_URL + 'groups.history', 'timeline', 'messages')
    GROUPS_INFO = (ROOT_URL + 'groups.info', None, None)
    GROUPS_INVITE = (ROOT_URL + 'groups.invite', None, None)
    GROUPS_KICK = (ROOT_URL + 'groups.kick', None, None)
    GROUPS_LEAVE = (ROOT_URL + 'groups.leave', None, None)
    GROUPS_LIST = (ROOT_URL + 'groups.list', None, None)
    GROUPS_MARK = (ROOT_URL + 'groups.mark', None, None)
    GROUPS_OPEN = (ROOT_URL + 'groups.open', None, None)
    GROUPS_RENAME = (ROOT_URL + 'groups.rename', None, None)
    GROUPS_REPLIES = (ROOT_URL + 'groups.replies', None, None)
    GROUPS_SET_PURPOSE = (ROOT_URL + 'groups.setPurpose', None, None)
    GROUPS_SET_TOPIC = (ROOT_URL + 'groups.setTopic', None, None)
    GROUPS_UNARCHIVE = (ROOT_URL + 'groups.unarchive', None, None)

    # im
    IM_CLOSE = (ROOT_URL + 'im.close', None, None)
    IM_HISTORY = (ROOT_URL + 'im.history', 'timeline', 'messages')
    IM_LIST = (ROOT_URL + 'im.list', None, None)
    IM_MARK = (ROOT_URL + 'im.mark', None, None)
    IM_OPEN = (ROOT_URL + 'im.open', None, None)
    IM_REPLIES = (ROOT_URL + 'im.replies', None, None)

    # mpim
    MPIM_CLOSE = (ROOT_URL + 'mpim.close', None, None)
    MPIM_HISTORY = (ROOT_URL + 'mpim.history', 'timeline', 'messages')
    MPIM_LIST = (ROOT_URL + 'mpim.list', None, None)
    MPIM_MARK = (ROOT_URL + 'mpim.mark', None, None)
    MPIM_OPEN = (ROOT_URL + 'mpim.open', None, None)
    MPIM_REPLIES = (ROOT_URL + 'mpim.replies', None, None)

    # oauth
    OAUTH_ACCESS = (ROOT_URL + 'oauth.access', None, None)
    OAUTH_TOKEN = (ROOT_URL + 'oauth.token', None, None)

    # pins
    PINS_ADD = (ROOT_URL + 'pins.add', None, None)
    PINS_LIST = (ROOT_URL + 'pins.list', None, None)
    PINS_REMOVE = (ROOT_URL + 'pins.remove', None, None)

    # reactions
    REACTIONS_ADD = (ROOT_URL + 'reactions.add', None, None)
    REACTIONS_GET = (ROOT_URL + 'reactions.get', None, None)
    REACTIONS_LIST = (ROOT_URL + 'reactions.list', 'page', 'items')
    REACTIONS_REMOVE = (ROOT_URL + 'reactions.remove', None, None)

    # reminders
    REMINDERS_ADD = (ROOT_URL + 'reminders.add', None, None)
    REMINDERS_COMPLETE = (ROOT_URL + 'reminders.complete', None, None)
    REMINDERS_DELETE = (ROOT_URL + 'reminders.delete', None, None)
    REMINDERS_INFO = (ROOT_URL + 'reminders.info', None, None)
    REMINDERS_LIsT = (ROOT_URL + 'reminders.list', None, None)

    # rtm
    RTM_CONNECT = (ROOT_URL + 'rtm.connect', None, None)
    RTM_START = (ROOT_URL + 'rtm.start', None, None)

    # search
    SEARCH_ALL = (ROOT_URL + 'search.all', 'page', 'messages')
    SEARCH_FILES = (ROOT_URL + 'search.files', 'page', 'files')
    SEARCH_MESSAGES = (ROOT_URL + 'search.messages', 'page', 'messages')

    # starts
    STARS_ADD = (ROOT_URL + 'stars.add', None, None)
    STARS_LIST = (ROOT_URL + 'stars.list', 'page', 'items')
    STARS_REMOVE = (ROOT_URL + 'stars.remove', None, None)

    # team
    TEAM_ACCESS_LOGS = (ROOT_URL + 'teams.accessLogs', None, None)
    TEAM_BILLABLE_INFO = (ROOT_URL + 'teams.billableInfo', None, None)
    TEAM_INFO = (ROOT_URL + 'teams.info', None, None)
    TEAM_INTEGRATION_LOGS = (ROOT_URL + 'teams.integrationLogs', None, None)

    # team profile
    TEAM_PROFILE_GET = (ROOT_URL + 'teams.profile.get', None, None)

    # usergroups
    USERGROUPS_CREATE = (ROOT_URL + 'usergroups.create', None, None)
    USERGROUPS_DISABLE = (ROOT_URL + 'usergroups.disable', None, None)
    USERGROUPS_ENABLE = (ROOT_URL + 'usergroups.enable', None, None)
    USERGROUPS_LIST = (ROOT_URL + 'usergroups.list', None, None)
    USERGROUPS_UPDATE = (ROOT_URL + 'usergroups.update', None, None)

    # usergroups users
    USERGROUPS_USERS_LIST = (ROOT_URL + 'usergroups.users.list', None, None)
    USERGROUPS_USERS_UPDATE = (ROOT_URL + 'usergroups.users.update', None, None)

    # users
    USERS_DELETE_PHOTO = (ROOT_URL + 'users.deletePhoto', None, None)
    USERS_GET_PRESENCE = (ROOT_URL + 'users.getPresence', None, None)
    USERS_IDENTITY = (ROOT_URL + 'users.identity', None, None)
    USERS_INFO = (ROOT_URL + 'users.info', None, None)
    USERS_LIST = (ROOT_URL + 'users.list', 'cursor', 'members')
    USERS_SET_ACTIVE = (ROOT_URL + 'users.setActive', None, None)
    USERS_SET_PHOTO = (ROOT_URL + 'users.setPhoto', None, None)
    USERS_SET_PRESENCE = (ROOT_URL + 'users.setPresence', None, None)

    # users profile
    USERS_PROFILE_GET = (ROOT_URL + 'users.profile.get', None, None)
    USERS_PROFILE_SET = (ROOT_URL + 'users.profile.set', None, None)



import aiohttp
import asyncio

from . import abc


class SlackAPI(abc.SlackAPI):
    """
    `aiohttp` implementation of :class:`slack.io.abc.SlackAPI`
    """
    def __init__(self, *, session, **kwargs):
        self._session = session
        super().__init__(**kwargs)

    async def _request(self, method, url, headers, body):
        async with self._session.request(method, url, headers=headers,
                                         data=body) as response:
            return response.status, await response.read(), response.headers

    async def _rtm(self, url):

        async with self._session.ws_connect(url) as ws:
            async for data in ws:
                if data.type == aiohttp.WSMsgType.TEXT:
                    yield data.data
                elif data.type == aiohttp.WSMsgType.CLOSED:
                    break
                elif data.type == aiohttp.WSMsgType.ERROR:
                    break

    async def sleep(self, seconds):
        await asyncio.sleep(seconds)



import abc
import json
import time
import logging

from .. import sansio, exceptions, events, methods

LOG = logging.getLogger(__name__)


class SlackAPI(abc.ABC):
    """
    :py:term:`abstract base class` abstracting the HTTP library used to call Slack API. Built with the functions of
    :mod:`slack.sansio`.

    Args:
        session: HTTP session
        token: Slack API token
        retry_when_rate_limit: Retry or raise an exception when rate limited
        headers: Default headers for all request

    Attributes:
        rate_limited: If rate limited timestamp when rate limit stop

    """
    def __init__(self, *, token, retry_when_rate_limit=True, headers=None):
        self._token = token
        self._headers = headers or {}
        self._retry_when_rate_limit = retry_when_rate_limit

        self.rate_limited = False

    @abc.abstractmethod
    async def _request(self, method, url, headers, body):
        return '', {}, b''

    @abc.abstractmethod
    async def _rtm(self, url):
        pass

    @abc.abstractmethod
    async def sleep(self, seconds):
        pass

    async def _make_query(self, url, data=None, headers=None):

        while self.rate_limited and self.rate_limited > int(time.time()):
            await self.sleep(1)

        try:
            url, body, headers = sansio.prepare_request(url=url, data=data, headers=headers,
                                                        global_headers=self._headers, token=self._token)
            status, body, headers = await self._request('POST', url, headers, body)
            response_data = sansio.decode_response(status, headers, body)
        except exceptions.RateLimited as rate_limited:
            if self._retry_when_rate_limit:
                raise
            else:
                LOG.warning('Rate limited ! Waiting for %s seconds', rate_limited.retry_after)
                self.rate_limited = int(time.time()) + rate_limited.retry_after
                return await self._make_query(url, data, headers)
        else:
            self.rate_limited = False
            return response_data

    async def query(self, url, data=None, headers=None):
        """
        Query the slack API

        Args:
            url: :class:`slack.methods` or url string
            data: JSON encodable MutableMapping
            headers: Custom headers

        Returns:
            dictionary of slack API response data

        """

        return await self._make_query(url, data, headers)

    async def iter(self, url, data=None, headers=None, *, limit=200, iterkey=None, itermode=None):
        """
        Iterate over a slack API method supporting pagination

        Args:
            url: :class:`slack.methods` or url string
            data: JSON encodable MutableMapping
            headers:
            limit: Maximum number of results to return per call.
            iterkey: Key in response data to iterate over (required for url string).
            itermode: Iteration mode (required for url string) (one of `cursor`, `page` or `timeline`)

        Returns:
            Async iterator over `response_data[key]`

        """
        itervalue = None
        while True:
            data, iterkey, itermode = sansio.prepare_iter_request(url, data, iterkey=iterkey, itermode=itermode,
                                                                  limit=limit, itervalue=itervalue)
            response_data = await self._make_query(url, data, headers)
            itervalue = sansio.decode_iter_request(response_data)
            for item in response_data[iterkey]:
                yield item

            if not itervalue:
                break

    async def rtm(self, url=None, bot_id=None):
        """
        Iterate over event from the RTM API

        Args:
            url: Websocket connection url
            bot_id: Connecting bot ID

        Returns:
            :class:`slack.events.Event` or :class:`slack.events.Message`

        """
        while True:
            if not bot_id:
                auth = await self.query(methods.AUTH_TEST)
                user_info = await self.query(methods.USERS_INFO, {'user': auth['user_id']})
                bot_id = user_info['user']['profile']['bot_id']
                LOG.info('BOT_ID is %s', bot_id)
            if not url:
                url = (await self.query(methods.RTM_CONNECT))['url']

            async for data in self._rtm(url):
                event = events.Event.from_rtm(json.loads(data))
                if sansio.need_reconnect(event):
                    break
                elif sansio.discard_event(event, bot_id):
                    if event['type'] == 'reconnect_url':
                        url = event['url']
                else:
                    yield event



import logging

from collections import defaultdict, MutableMapping
from . import exceptions

LOG = logging.getLogger(__name__)


class Action(MutableMapping):
    """
    MutableMapping representing a response to an interactive message.

    Args:
        raw_action: Decoded body of the HTTP request
        verification_token: Slack verification token used to verify the request came from slack
        team_id: Verify the event is for the correct team
    Raises:
        :class:`slack.exceptions.FailedVerification`: when `verification_token` or `team_id` does not match the
                                                      incoming event's
    """

    def __init__(self, raw_action, verification_token=None, team_id=None):
        self.action = raw_action

        if verification_token and self.action['token'] != verification_token:
            raise exceptions.FailedVerification(self.action['token'], self.action['team_id'])

        if team_id and self.action['team_id'] != team_id:
            raise exceptions.FailedVerification(self.action['token'], self.action['team_id'])

    def __getitem__(self, item):
        return self.action[item]

    def __setitem__(self, key, value):
        self.action[key] = value

    def __delitem__(self, key):
        del self.action[key]

    def __iter__(self):
        return iter(self.action)

    def __len__(self):
        return len(self.action)

    def __repr__(self):
        return str(self.action)


class Router:
    """
    When creating a slack applications you can only set one action url. This provide a routing mechanism for the
    incoming actions, based on their `callback_id`, to one or more handlers.
    """
    def __init__(self):
        self._routes = defaultdict(dict)

    def register(self, callback_id, handler):
        """
        Register a new handler for a specific :class:`slack.actions.Action` `callback_id`.

        Args:
            callback_id: Callback_id the handler is interested in
            handler: Callback
        """
        LOG.info('Registering %s to %s', callback_id, handler.__name__)
        self._routes[callback_id].append(handler)

    def dispatch(self, action):
        """
        Yields handlers matching the incoming :class:`slack.actions.Action` `callback_id`.

        Args:
            action: :class:`slack.actions.Action`

        Yields:
            handler
        """
        LOG.debug('Dispatching action %s', action['callback_id'])

        for callback in self._routes.get(action['callback_id']):
            yield callback



"""
Collection of functions for sending and decoding request to or from the slack API
"""

import cgi
import json
import logging

from . import exceptions, events, methods, ROOT_URL

LOG = logging.getLogger(__name__)

RECONNECT_EVENTS = ('team_migration_started', 'goodbye')
"""Events type preceding a disconnection"""

SKIP_EVENTS = ('reconnect_url', )
"""Events that do not need to be dispatched"""

ITERMODE = ('cursor', 'page', 'timeline')
"""Supported pagination mode"""


def raise_for_status(status, headers, data):
    """
    Check request response status

    Args:
        status: Response status
        headers: Response headers
        data: Response data

    Raises:
        :class:`slack.exceptions.RateLimited`: For 429 status code
        :class:`slack.exceptions:HTTPException`:
    """
    if status != 200:
        if status == 429:
            try:
                retry_after = int(headers.get('Retry-After', 1))
            except ValueError:
                retry_after = 1
            raise exceptions.RateLimited(retry_after, data.get('error', 'ratelimited'), status, headers, data)
        else:
            raise exceptions.HTTPException(status, headers, data)


def raise_for_api_error(headers, data):
    """
    Check request response for Slack API error

    Args:
        headers: Response headers
        data: Response data

    Raises:
        :class:`slack.exceptions.SlackAPIError`
    """
    if not data['ok']:
        raise exceptions.SlackAPIError(data.get('error', 'unknow_error'), headers, data)

    if 'warning' in data:
        LOG.warning(data['warning'])


def decode_body(headers, body):
    """
    Decode the response body

    For 'application/json' content-type load the body as a dictionary

    Args:
        headers: Response headers
        body: Response body

    Returns:
        decoded body
    """
    type_, encoding = parse_content_type(headers)
    decoded_body = body.decode(encoding)

    if type_ == 'application/json':
        return json.loads(decoded_body)
    else:
        return decoded_body


def parse_content_type(headers):
    """
    Find content-type and encoding of the response

    Args:
        headers: Response headers

    Returns:
        :py:class:`tuple` (content-type, encoding)
    """
    content_type = headers.get('content-type')
    if not content_type:
        return None, 'utf-8'
    else:
        type_, parameters = cgi.parse_header(content_type)
        encoding = parameters.get("charset", "utf-8")
        return type_, encoding


def prepare_request(url, data, headers, global_headers, token):
    """
    Prepare outgoing request

    Create url, headers, add token to the body if needed json encode it

    Args:
        url: :class:`slack.methods` item or string of url
        data: Outgoing data
        headers: Custom headers
        global_headers: Global headers
        token: Slack API token

    Returns:
        :py:class:`tuple` (url, body, headers)
    """
    if isinstance(data, events.Message):
        data = data.serialize()

    if isinstance(url, methods):
        url = url.value[0]
    elif not url.startswith(ROOT_URL):
        url = ROOT_URL + url

    if not headers:
        headers = {**global_headers}
    else:
        headers = {**global_headers, **headers}

    if not data:
        data = {'token': token}
    elif 'token' not in data:
        data['token'] = token

    if url.startswith('https://hooks.slack.com'):
        body = json.dumps(data)
    else:
        body = data

    return url, body, headers


def decode_response(status, headers, body):
    """
    Decode incoming response

    Args:
        status: Response status
        headers: Response headers
        body: Response body

    Returns:
        Response data
    """
    data = decode_body(headers, body)
    raise_for_status(status, headers, data)
    raise_for_api_error(headers, data)

    return data


def find_iteration(url, itermode=None, iterkey=None):
    """
    Find iteration mode and iteration key for a given :class:`slack.methods`

    Args:
        url: :class:`slack.methods` or string url
        itermode: Custom iteration mode
        iterkey: Custom iteration key

    Returns:
        :py:class:`tuple` (itermode, iterkey)
    """
    if isinstance(url, methods):
        if not itermode:
            itermode = url.value[1]
        if not iterkey:
            iterkey = url.value[2]

    if not iterkey or not itermode:
        raise ValueError('Iteration not supported for: {}'.format(url))
    elif itermode not in ITERMODE:
        raise ValueError('Iteration not supported for: {}'.format(itermode))

    return itermode, iterkey


def prepare_iter_request(url, data, *, iterkey=None, itermode=None, limit=200, itervalue=None):
    """
    Prepare outgoing iteration request

    Args:
        url: :class:`slack.methods` item or string of url
        data: Outgoing data
        limit: Maximum number of results to return per call.
        iterkey: Key in response data to iterate over (required for url string).
        itermode: Iteration mode (required for url string) (one of `cursor`, `page` or `timeline`)
        itervalue: Value for current iteration (cursor hash, page or timestamp depending on the itermode)
    Returns:
        :py:class:`tuple` (data, iterkey, itermode)
    """
    itermode, iterkey = find_iteration(url, itermode, iterkey)

    if not data:
        data = {}

    if itermode == 'cursor':
        data['limit'] = limit
        if itervalue:
            data['cursor'] = itervalue
    elif itermode == 'page':
        data['count'] = limit
        if itervalue:
            data['page'] = itervalue
    elif itermode == 'timeline':
        data['count'] = limit
        if itervalue:
            data['latest'] = itervalue
    else:
        raise NotImplementedError('Unknown itermode: %s', itermode)

    return data, iterkey, itermode


def decode_iter_request(data):
    """
    Decode incoming response from an iteration request

    Args:
        data: Response data

    Returns:
        Next itervalue
    """
    if 'response_metadata' in data:
        return data['response_metadata'].get('next_cursor')
    elif 'paging' in data:
        current_page = int(data['paging'].get('page', 1))
        max_page = int(data['paging'].get('pages', 1))

        if current_page < max_page:
            return current_page + 1
    elif 'has_more' in data and data['has_more'] and 'latest' in data:
        return data['messages'][-1]['ts']


def discard_event(event, bot_id=None):
    """
    Check if the incoming event needs to be discarded

    Args:
        event: Incoming :class:`slack.events.Event`
        bot_id: Id of connected bot

    Returns:
        boolean
    """
    if event['type'] in SKIP_EVENTS:
        return True
    elif bot_id and isinstance(event, events.Message):
        if event.get('bot_id') == bot_id:
            LOG.debug('Ignoring event: %s', event)
            return True
        elif 'message' in event and event['message'].get('bot_id') == bot_id:
            LOG.debug('Ignoring event: %s', event)
            return True

    return False


def need_reconnect(event):
    """
    Check if RTM needs reconnecting

    Args:
        event: Incoming :class:`slack.events.Event`

    Returns:
        boolean
    """
    if event['type'] in RECONNECT_EVENTS:
        return True
    else:
        return False



import pathlib
import setuptools

LONG_DESCRIPTION = pathlib.Path('README.rst').read_text('utf-8')

requires = {
    'install': [],
    'setup': [],
    'tests': ['flake8'],
    'full': set(),
    'doc': {'sphinx', 'sphinxcontrib-asyncio', 'sphinxcontrib-napoleon'},
    'dev': {'tox'},
    'requests': ['requests', 'websocket-client'],
    'aiohttp': ['aiohttp'],
    'curio': ['curio', 'asks'],
    'trio': ['trio', 'asks'],
    'treq': ['treq'],
}

requires['dev'].update(*requires.values())
requires['full'].update(requires['requests'], requires['aiohttp'], requires['curio'], requires['trio'],
                        requires['treq'])


def find_version():
    with open("slack/__version__.py") as f:
        version = f.readlines()[-1].split('=')[-1].strip().strip("'").strip('"')
        if not version:
            raise RuntimeError('No version found')

    return version


setuptools.setup(
    name='slack-sansio',
    long_description=LONG_DESCRIPTION,
    description='(a)sync Slack API library',
    keywords=[
        'bot',
        'slack',
        'api',
        'sans-io',
        'async'
    ],
    packages=setuptools.find_packages(),
    zip_safe=True,
    install_requires=requires['install'],
    setup_requires=requires['setup'],
    tests_require=requires['tests'],
    extras_require=requires,
    # python_requires='>=3.6',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
    ],
    author='Ovv',
    author_email='contact@ovv.wtf',
    license='MIT',
    url='https://github.com/pyslackers/slack-sansio',
    version=find_version(),
)
